// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

// Performs Baum-Welch training with LM and output strings.

#include <unistd.h>

#include <cstdlib>
#include <cstring>
#include <ctime>

#include <memory>
#include <string>
using std::string;

#include <fst/flags.h>

#include <baumwelch/baumwelchscript.h>

DEFINE_bool(decipherment, false,
            "Use decipherment construction; i.e., plaintext is a WFSA rather "
            "than a FAR");
DEFINE_string(expectation_table, "state_ilabel",
              "Expectation table, one of: \"global\", \"state\", \"ilabel\", "
              "\"state_ilabel\"");
DEFINE_int32(max_iters, fst::kMaxIters,
             "Maximum number of iterations to perform");
DEFINE_bool(flat_start, true, "Perform one round of flat start training?");
DEFINE_int32(random_starts, fst::kRandomStarts,
             "Number of random starts to perform");
DEFINE_bool(remove_zero_arcs, true, "Should zero arcs be removed?");
DEFINE_double(delta, fst::kDelta, "Comparison/quantization delta");
DEFINE_int32(seed, time(nullptr) + getpid(), "Random seed");

int main(int argc, char **argv) {
  namespace s = fst::script;
  using fst::BaumWelchTrainOptions;
  using fst::ExpectationTableType;
  using fst::script::FarReaderClass;
  using fst::script::FstClass;
  using fst::script::MutableFstClass;

  string usage =
      "Performs Baum-Welch training with an LM and FAR inputs "
      "\n\n  Usage: ";
  usage += argv[0];
  usage += " plaintext.f(ar|st) ciphertext.far channel.fst [out.fst]\n";

  std::set_new_handler(FailedNewHandler);
  SET_FLAGS(usage.c_str(), &argc, &argv, true);
  if (argc < 4 || argc > 5) {
    ShowUsage();
    return 1;
  }

  const string plaintext_name = strcmp(argv[1], "-") != 0 ? argv[1] : "";
  const string ciphertext_name = strcmp(argv[2], "-") != 0 ? argv[2] : "";
  const string channel_name = strcmp(argv[3], "-") != 0 ? argv[3] : "";
  const string out_name = argc > 4 ? argv[4] : "";

  if (plaintext_name.empty() &&
      (ciphertext_name.empty() || channel_name.empty())) {
    LOG(ERROR) << argv[0] << ": Can't take more than one input from standard "
               << "input";
    return 1;
  }
  if (ciphertext_name.empty() && channel_name.empty()) {
    LOG(ERROR) << argv[0] << ": Can't take more than one input from standard "
               << "input";
    return 1;
  }

  const std::unique_ptr<FarReaderClass> ciphertext(
      FarReaderClass::Open(ciphertext_name));
  if (!ciphertext) return 1;

  const std::unique_ptr<MutableFstClass> channel(
      MutableFstClass::Read(channel_name));
  if (!channel) return 1;

  ExpectationTableType etype;
  if (!s::GetExpectationTableType(FLAGS_expectation_table, &etype)) {
    LOG(ERROR) << argv[0] << ": Unknown or unsupported expectation table type: "
               << FLAGS_expectation_table;
    return 1;
  }

  const BaumWelchTrainOptions opts(FLAGS_max_iters, FLAGS_flat_start,
                                   FLAGS_random_starts, FLAGS_remove_zero_arcs,
                                   FLAGS_delta);

  srand(FLAGS_seed);

  if (!FLAGS_decipherment) {
    const std::unique_ptr<FarReaderClass> plaintext(
        FarReaderClass::Open(plaintext_name));
    if (!plaintext) return 1;

    if (!TrainBaumWelch(plaintext.get(), ciphertext.get(), channel.get(), etype,
                        opts)) {
      LOG(WARNING) << argv[0] << ": Training did not converge";
    }
  } else {
    const std::unique_ptr<const FstClass> plaintext(
        FstClass::Read(plaintext_name));
    if (!plaintext) return 1;

    if (!TrainBaumWelch(*plaintext, ciphertext.get(), channel.get(), etype,
                        opts)) {
      LOG(WARNING) << argv[0] << ": Training did not converge";
    }
  }

  channel->Write(out_name);

  return 0;
}

