// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#ifndef BAUMWELCH_BAUMWELCH_H_
#define BAUMWELCH_BAUMWELCH_H_

// Expectation-maximization-style training for weighted finite-state
// transducers.
//
// In this model, the weights of a "channel" FST (a discrete-alphabet HMM) are
// optimized to maximize the joint probability with respect to some other data
// (also represented as one or more FSTs).
//
// The training data consist of one an LM-like input FST and one more more
// output FSTs. This is useful for "unsupervised" training (what Kevin Knight
// and colleagues call "decipherment") in the sense that it will learn a
// conditional probability model that maximizes the probability of the outputs
// with respect to the LM.
//
// A Likelihood() method returns the global likelihood on the previous
// iteration. Ciphertext FSTs are provided to the model using multiple
// invocations of CollectExpectations(), and the Maximize() method writes the
// collected counts back into the channel model FST.
//
// Note that while final weights contribute to the calculations of iteration
// likelihood, they are treated as fixed and not effected by maximization.
//
// Arcs not traversed during training are assigned a semiring Zero weight and
// these arcs can be removed using the RemoveZeroArcs method. Assuming the
// training corpus is fixed, this need only be applied after the first
// iteration.
//
// If the channel model contains any multiarcs (arcs with the same source,
// destination, and labeling---but possibly with different weights) these can
// be combined using the arc-sum mapper ahead of time for efficiency.
//
// By convention, operations in this library that work with FarReader input
// reset the FAR to its initial position upon completion.

#include <utility>
#include <vector>

#include <fst/types.h>
#include <fst/log.h>
#include <fst/compat.h>
#include <fst/extensions/far/far.h>
#include <fst/fstlib.h>
#include <baumwelch/expectation-table.h>
#include <baumwelch/log-math.h>

namespace fst {

constexpr int32 kMaxIters = 50;
constexpr int32 kRandomStarts = 10;

enum ExpectationTableType { GLOBAL, STATE, ILABEL, STATE_ILABEL };

// Forward-declaration.
template <class Arc, class ExpectationTable>
class BaumWelch;

// Struct for training options.
struct BaumWelchTrainOptions {
  explicit BaumWelchTrainOptions(int32 max_iters = kMaxIters,
                                 bool flat_start = true,
                                 int32 random_starts = kRandomStarts,
                                 bool remove_zero_arcs = true,
                                 float delta = kDelta)
      : max_iters(max_iters),
        flat_start(flat_start),
        random_starts(random_starts),
        remove_zero_arcs(remove_zero_arcs),
        delta(delta) {}

  // Maximum number of iterations to perform; if 0, the channel is just
  // normalized and no other work is done.
  int32 max_iters;
  // Perform one round of flat start training (i.e., as the 0th "restart")?
  bool flat_start;
  // Number of random starts to perform; if 0, random start training is not
  // performed.
  int32 random_starts;
  // Should zero-weight arcs in the cascade be removed from the channel model
  // before training?
  bool remove_zero_arcs;
  // Comparison/quantization delta used to determine convergence.
  float delta;
};

namespace internal {

// Objects and implementations used in the BaumWelch class.

// Comparison functor used to arc-sort the channel model for efficient
// composition. Under certain conditions (including the obvious one where the
// FST is an acceptor), an FST so sorted will be both input- and output-sorted.
template <class Arc>
class OILabelCompare {
 public:
  bool operator()(const Arc &arc1, const Arc &arc2) const {
    if (arc1.olabel < arc2.olabel) return true;
    return (arc1.olabel == arc2.olabel) ? arc1.ilabel < arc2.ilabel : false;
  }

  uint64 Properties(uint64 props) const {
    return (props & kArcSortProperties) | kOLabelSorted |
           (props & kAcceptor ? kILabelSorted : 0);
  }
};

// Properties of the channel model preserved by operations below.
constexpr uint64 kBaumWelchChannelProperties = kILabelSorted | kOLabelSorted;

// Removes arcs with zero weights in the channel model, and connects the result
// if necessary; should only be called only immediately after expectations have
// been collected.
template <class Arc>
void RemoveZeroArcs(MutableFst<Arc> *channel) {
  using Weight = typename Arc::Weight;
  bool mutation = false;
  const auto props = channel->Properties(kBaumWelchChannelProperties, false);
  const auto dead_state = channel->NumStates();
  for (StateIterator<MutableFst<Arc>> siter(*channel); !siter.Done();
       siter.Next()) {
    for (MutableArcIterator<MutableFst<Arc>> aiter(channel, siter.Value());
         !aiter.Done(); aiter.Next()) {
      const auto &arc = aiter.Value();
      if (arc.weight == Weight::Zero()) {
        mutation = true;
        aiter.SetValue(Arc(arc.ilabel, arc.olabel, arc.weight, dead_state));
      }
    }
  }
  if (mutation) {
    channel->AddState();
    Connect(channel);
    channel->SetProperties(props, kBaumWelchChannelProperties);
  }
}

// Assigns random weights sampled from (0, 1] to arcs; final weights are not
// affected.
template <class Arc>
void RandomizeArcWeights(MutableFst<Arc> *channel) {
  using Weight = typename Arc::Weight;
  using ValueType = typename Weight::ValueType;
  const auto props = channel->Properties(kBaumWelchChannelProperties, false);
  static const LogUniformGenerator<ValueType> generate;
  for (StateIterator<MutableFst<Arc>> siter(*channel); !siter.Done();
       siter.Next()) {
    for (MutableArcIterator<MutableFst<Arc>> aiter(channel, siter.Value());
         !aiter.Done(); aiter.Next()) {
      const auto &arc = aiter.Value();
      const Arc narc(arc.ilabel, arc.olabel, Weight(generate()), arc.nextstate);
      aiter.SetValue(narc);
    }
  }
  channel->SetProperties(props, kBaumWelchChannelProperties);
}

// Function for computing shortest distances as part of the E-step.
template <class Arc>
bool GetShortestDistances(const Fst<Arc> &ico,
                          std::vector<typename Arc::Weight> *alpha,
                          std::vector<typename Arc::Weight> *beta) {
  using Weight = typename Arc::Weight;
  // Computes backwards weights.
  ShortestDistance(ico, beta, true);
  if (beta->empty()) {
    LOG(WARNING) << "CollectExpectations: No non-zero paths";
    return false;
  }
  // Computes forward weights.
  ShortestDistance(ico, alpha, false);
  return true;
}

// Class for storing forward and backwards weights.
template <class Arc>
class ForwardBackward {
 public:
  using StateId = typename Arc::StateId;
  using Weight = typename Arc::Weight;

  explicit ForwardBackward(const Fst<Arc> &ico) {
    // Backwards weights.
    ShortestDistance(ico, &beta_, true);
    if (beta_.empty()) LOG(WARNING) << "ShortestDistances: Non non-zero paths";
    // Forward weights.
    ShortestDistance(ico, &alpha_, false);
    // FIXME(kbg): Add debug
  }

  const Weight &Alpha(StateId s) const {
    return ForwardBackward<Arc>::GetWeightOrZero(alpha_, s);
  }

  const Weight &Beta(StateId s) const {
    return ForwardBackward<Arc>::GetWeightOrZero(beta_, s);
  }

 private:
  // This static method is used to get a weight from a weight vector produced
  // by ShortestDistance; it returns semiring Zero in the case the state ID
  // exceeds the number of states in the weight vector; it also checks
  // weight membership.
  static const Weight &GetWeightOrZero(const std::vector<Weight> &weights,
                                       StateId s) {
    if (s >= weights.size()) {
      return Weight::Zero();
    } else {
      const auto &weight = weights[s];
      CHECK(weight.Member());
      return weight;
    }
  }

  std::vector<Weight> alpha_;
  std::vector<Weight> beta_;
};

// Single observation during the E-step.
template <class ExpectationTable>
bool CollectExpectations(const typename ExpectationTable::Cascade &cascade,
                         ExpectationTable *e_table) {
  using Arc = typename ExpectationTable::Arc;
  using Weight = typename Arc::Weight;
  // Populates shortest distance vectors.
  const auto &ico = cascade.GetFst();
  const ForwardBackward<Arc> fb(ico);
  // Collects likelihood for observation using the final weight.
  const auto start = ico.Start();
  if (start == kNoStateId) {
    LOG(WARNING) << "CollectExpectations: Composition failure";
    return false;
  }
  e_table->CollectLikelihood(fb.Beta(start).Value());
  // Actually collects expectations for observation.
  for (StateIterator<Fst<Arc>> siter(ico); !siter.Done(); siter.Next()) {
    const auto state = siter.Value();
    // Non-coaccessible source state.
    if (fb.Beta(state) == Weight::Zero()) continue;
    const auto ch_state = cascade.ChannelState(state);
    const auto &aref = fb.Alpha(state);
    for (ArcIterator<Fst<Arc>> aiter(ico, state); !aiter.Done(); aiter.Next()) {
      const auto &arc = aiter.Value();
      const auto &bref = fb.Beta(arc.nextstate);
      // Non-coaccesssible destination state.
      if (bref == Weight::Zero()) continue;
      const auto e_weight =
          internal::LogTimes(aref.Value(), arc.weight.Value(), bref.Value());
      const auto ch_nextstate = cascade.ChannelState(arc.nextstate);
      e_table->CollectExpectation(ch_state, arc.ilabel, arc.olabel, e_weight,
                                  ch_nextstate);
    }
  }
  return true;
}

// The M-step.
template <class ExpectationTable>
void Maximize(MutableFst<typename ExpectationTable::Arc> *channel,
              ExpectationTable *e_table) {
  using Arc = typename ExpectationTable::Arc;
  using Weight = typename Arc::Weight;
  const auto props = channel->Properties(kBaumWelchChannelProperties, false);
  for (StateIterator<MutableFst<Arc>> siter(*channel); !siter.Done();
       siter.Next()) {
    const auto state = siter.Value();
    for (MutableArcIterator<MutableFst<Arc>> aiter(channel, state);
         !aiter.Done(); aiter.Next()) {
      // TODO(kbg): Add a normalizing factor version of this to prevent
      // vanishing expectations.
      auto arc = aiter.Value();
      arc.weight = Weight(e_table->Maximize(state, arc));
      aiter.SetValue(arc);
    }
  }
  channel->SetProperties(props, internal::kBaumWelchChannelProperties);
}

// Round of multi-iteration training starting from the current model.
template <class Arc, class ExpectationTable, class DataPair>
bool TrainFlat(DataPair *datapair, BaumWelch<Arc, ExpectationTable> *baumwelch,
               int32 max_iters = kMaxIters, float delta = kDelta) {
  int32 iter = 0;
  // Zeroth iteration, which is just an MLE step in the case the model is
  // unnormalized. As this is often the case we don't bother to log the
  // likelihood.
  if (!baumwelch->Iteration(datapair->first, datapair->second)) return false;
  // First true iteration.
  if (max_iters < 1) return false;
  if (!baumwelch->Iteration(datapair->first, datapair->second)) return false;
  ++iter;
  auto old_likelihood = baumwelch->Likelihood();
  VLOG(1) << "Iteration " << iter << " likelihood: " << old_likelihood;
  while (iter < max_iters) {
    ++iter;
    if (!baumwelch->Iteration(datapair->first, datapair->second)) return false;
    const auto likelihood = baumwelch->Likelihood();
    VLOG(1) << "Iteration " << iter << " likelihood: " << likelihood;
    if (LogApproxEqual(old_likelihood, likelihood, delta)) {
      VLOG(1) << "Convergence (" << iter << " iterations)";
      return true;
    }
    old_likelihood = likelihood;
  }
  VLOG(1) << "Did not converge (" << max_iters << " iterations)";
  return false;
}

// Full training regime, consisting of:
//
// 0. construction of the Baum-Welch class instance
// 1. optional removal of zero arcs
// 2. zero or one round of flat start training
// 3. zero or more rounds of random start training
// 4. destruction of the Baum-Welch class instance
template <class Arc, class ExpectationTable, class DataPair>
bool Train(DataPair *datapair, BaumWelch<Arc, ExpectationTable> *baumwelch,
           const BaumWelchTrainOptions &opts = BaumWelchTrainOptions()) {
  bool converged = false;
  if (opts.remove_zero_arcs) {
    VLOG(1) << "Removing zero arcs";
    if (!baumwelch->RemoveZeroArcs(datapair->first, datapair->second))
      return false;
  }
  int32 random_start = 0;
  if (opts.flat_start) {
    VLOG(1) << "Training (flat start)";
    converged = TrainFlat(datapair, baumwelch, opts.max_iters, opts.delta);
  } else if (opts.random_starts) {
    // In the case we don't do a flat start, we perform one iteration of random
    // training on the Baum-Welch object passed and treat it as that against
    // which we compare.
    VLOG(1) << "Training (random start 1)";
    baumwelch->RandomizeArcWeights();
    converged = TrainFlat(datapair, baumwelch, opts.max_iters, opts.delta);
    ++random_start;
  }
  while (random_start < opts.random_starts) {
    ++random_start;
    VLOG(1) << "Training (random start " << random_start << ")";
    BaumWelch<Arc, ExpectationTable> random(*baumwelch);
    random.RandomizeArcWeights();
    const bool random_converged =
        TrainFlat(datapair, &random, opts.max_iters, opts.delta);
    if (baumwelch->IsLessLikelyThan(random)) {
      VLOG(1) << "New best model (old likelihood: " << baumwelch->Likelihood()
              << "; new likelihood: " << random.Likelihood() << ")";
      *baumwelch = random;
      converged = random_converged;
    }
  }
  return converged;
}

// Helper function for users who don't require the low-level control provided
// by the BaumWelch class.
//
// Plaintext and ciphertext are stored in a std::pair to help with template
// parameter resolution.
template <class Arc, class DataPair>
bool TrainBaumWelch(
    DataPair *datapair, MutableFst<Arc> *channel,
    ExpectationTableType etype = STATE_ILABEL,
    const BaumWelchTrainOptions &opts = BaumWelchTrainOptions()) {
  switch (etype) {
    case GLOBAL: {
      BaumWelch<Arc, GlobalExpectationTable<Arc>> baumwelch(*channel);
      bool converged = Train(datapair, &baumwelch, opts);
      *channel = baumwelch.Channel();
      return converged;
    }
    case STATE: {
      BaumWelch<Arc, StateExpectationTable<Arc>> baumwelch(*channel);
      bool converged = Train(datapair, &baumwelch, opts);
      *channel = baumwelch.Channel();
      return converged;
    }
    case ILABEL: {
      BaumWelch<Arc, ILabelExpectationTable<Arc>> baumwelch(*channel);
      bool converged = Train(datapair, &baumwelch, opts);
      *channel = baumwelch.Channel();
      return converged;
    }
    case STATE_ILABEL: {
      BaumWelch<Arc, StateILabelExpectationTable<Arc>> baumwelch(*channel);
      bool converged = Train(datapair, &baumwelch, opts);
      *channel = baumwelch.Channel();
      return converged;
    }
    default: {
      FSTERROR() << "TrainBaumWelch: Unknown expectation table type: " << etype;
      return false;
    }
  }
}

}  // namespace internal

// Actual Baum-Welch trainer class.

template <class Arc, class ExpectationTable>
class BaumWelch {
 public:
  using ValueType = typename Arc::Weight::ValueType;

  explicit BaumWelch(const Fst<Arc> &channel)
      : channel_(channel), e_table_(channel_) {
    // O-sorts the channel (and tries to also i-sort it too).
    ArcSort(&channel_, internal::OILabelCompare<Arc>());
    // Computes the i-sort property.
    channel_.Properties(kILabelSorted, true);
  }

  BaumWelch &operator=(const BaumWelch &other) {
    channel_ = other.channel_;
    // NB: This doesn't really copy the expectations: just the table sizing.
    e_table_ = other.e_table_;
    return *this;
  }

  const VectorFst<Arc> &Channel() const { return channel_; }

  ValueType Likelihood() const { return e_table_.Likelihood(); }

  // Compares two models on negative log-likelihood; if true,
  // the current model is less likely than the argument model.
  bool IsLessLikelyThan(const BaumWelch &other) {
    return Likelihood() > other.Likelihood();
  }

  void RandomizeArcWeights() { internal::RandomizeArcWeights(&channel_); }

  // Removes unvisited arcs (pair construction).
  bool RemoveZeroArcs(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext) {
    if (!Iteration(plaintext, ciphertext)) return false;
    internal::RemoveZeroArcs(&channel_);
    return true;
  }

  // Removes unvisited arcs (decipherment construction).
  bool RemoveZeroArcs(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext) {
    if (!Iteration(plaintext, ciphertext)) return false;
    internal::RemoveZeroArcs(&channel_);
    return true;
  }

  // Training methods.

  // Single observation during the E-step.
  bool CollectExpectations(const Fst<Arc> &plaintext,
                           const Fst<Arc> &ciphertext) {
    using Cascade = typename ExpectationTable::Cascade;
    const Cascade cascade(plaintext, ciphertext, channel_);
    return internal::CollectExpectations(cascade, &e_table_);
  }

  // Entire E-step (pair construction).
  bool CollectExpectations(FarReader<Arc> *plaintext,
                           FarReader<Arc> *ciphertext) {
    bool success = true;
    // In the case they both do not contain the same number of FARs, we
    // terminate early.
    for (; !(plaintext->Done() || ciphertext->Done());
         plaintext->Next(), ciphertext->Next()) {
      if (!(success = CollectExpectations(*plaintext->GetFst(),
                                          *ciphertext->GetFst()))) {
        break;
      }
    }
    plaintext->Reset();
    ciphertext->Reset();
    return success;
  }

  // Entire E-step (decipherment construction).
  bool CollectExpectations(const Fst<Arc> &plaintext,
                           FarReader<Arc> *ciphertext) {
    bool success = true;
    for (; !ciphertext->Done(); ciphertext->Next()) {
      if (!(success = CollectExpectations(plaintext, *ciphertext->GetFst()))) {
        break;
      }
    }
    ciphertext->Reset();
    return success;
  }

  // Entire M-step.
  void Maximize() { internal::Maximize(&channel_, &e_table_); }

  // Entire EM iteration (pair construction).
  bool Iteration(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext) {
    Reset();
    if (!CollectExpectations(plaintext, ciphertext)) return false;
    Maximize();
    return true;
  }

  // Entire EM iteration (decipherment construction).
  bool Iteration(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext) {
    Reset();
    if (!CollectExpectations(plaintext, ciphertext)) return false;
    Maximize();
    return true;
  }

  // Round of multi-iteration training starting from the current model (pair
  // construction).
  bool TrainFlat(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
                 int32 max_iters = kMaxIters, float delta = kDelta) {
    // This is just a trick to help with template resolution.
    std::pair<FarReader<Arc> *, FarReader<Arc> *> datapair(plaintext,
                                                           ciphertext);
    return internal::TrainFlat(&datapair, this, max_iters, delta);
  }

  // Round of multi-iteration training starting from the current model (LM
  // construction).
  bool TrainFlat(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
                 int max_iters = kMaxIters, float delta = kDelta) {
    // This is just a trick to help with template resolution.
    std::pair<const Fst<Arc> &, FarReader<Arc> *> datapair(plaintext,
                                                           ciphertext);
    return internal::TrainFlat(&datapair, this, max_iters, delta);
  }

  // Round of multi-iterations training starting from a randomized model
  // (pair construction).
  bool TrainRandom(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
                   int max_iters = kMaxIters, float delta = kDelta) {
    RandomizeArcWeights();
    return TrainFlat(plaintext, ciphertext, max_iters, delta);
  }

  // Round of multi-iteration training starting from a randomized model (LM
  // construction).
  bool TrainRandom(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
                   int max_iters = kMaxIters, float delta = kDelta) {
    RandomizeArcWeights();
    return TrainFlat(plaintext, ciphertext, max_iters, delta);
  }

 protected:
  // Called immediately before starting a new iteration.
  void Reset() { e_table_.Reset(); }

 private:
  VectorFst<Arc> channel_;
  ExpectationTable e_table_;
};

// Helper functions for users who don't require the low-level control provided
// by the BaumWelch class.

// Baum-Welch training (pair construction).
template <class Arc>
bool TrainBaumWelch(
    FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel, ExpectationTableType etype = STATE_ILABEL,
    const BaumWelchTrainOptions &opts = BaumWelchTrainOptions()) {
  std::pair<FarReader<Arc> *, FarReader<Arc> *> datapair(plaintext, ciphertext);
  return internal::TrainBaumWelch(&datapair, channel, etype, opts);
}

// Baum-Welch training (decipherment construction).
template <class Arc>
bool TrainBaumWelch(
    const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel, ExpectationTableType etype = STATE_ILABEL,
    const BaumWelchTrainOptions &opts = BaumWelchTrainOptions()) {
  std::pair<const Fst<Arc> &, FarReader<Arc> *> datapair(plaintext, ciphertext);
  return internal::TrainBaumWelch(&datapair, channel, etype, opts);
}

}  // namespace fst

#endif  // BAUMWELCH_BAUMWELCH_H_

